/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Change the MEP of an exchange. Use the MEP defined in the Consumes element.
 * <br>
 * Depending on the MEP, some response can be lost. TODO sendAsync not managed
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * 
 */
public class Bridge extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.
     * component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
     */
    public void process(Exchange exchange, ExchangeContext context) throws PEtALSCDKException,
            MessagingException {
        Exchange originalExchange = exchange;
        List<Consumes> consumesList = context.getSUConsumes(originalExchange.getEndpoint());

        if (consumesList.size() == 1) {
            Consumes consumes = consumesList.get(0);

            // create the exchange with the new MEP to send to the next
            // service
            MEPConstants mep = PatternHelper.retrievePattern(consumes.getMep());
            Exchange newExchange = PatternHelper.createExchange(consumes, mep, originalExchange
                    .getInMessage(), context);

            // send the exchange to the next service
            logger.info("original M.E.P. : " + originalExchange.getPattern());
            logger.info("new      M.E.P. : " + newExchange.getPattern());
            logSend(consumes, exchange);
            context.sendSync(newExchange);

            // process the MEP mapping
            processMEPMapping(originalExchange, newExchange, context);

            // close the forwarded exchange if active
            if (newExchange.isActiveStatus()) {
                closeForwardedExchange(originalExchange, newExchange, context);
            }
        } else {
            throw new PEtALSCDKException(
                    "The Service Unit must define an unique Consumes sections for the pattern "
                            + this.getPatternName());
        }
    }

    /**
     * Process the mapping between the original and new MEP : <li>If an error
     * has been received, set it on the original MEP.</li> <li>Depending on the
     * original MEP, copy received Fault or Out message to the original
     * Exchange, if any; otherwise copy an empty Out message.</li>
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    protected void processMEPMapping(Exchange originalExchange, Exchange newExchange,
            ExchangeContext context) throws MessagingException, PEtALSCDKException {

        logger
                .finest("process the response according to the MEP mapping between the consumer and the provider.");
        // copy back the error from the forwarded exchange if any
        if (newExchange.isErrorStatus()) {
            setErrorOnOriginalExchange(originalExchange, newExchange, context);
        } else {
            // MEP mapping
            if (originalExchange.isInOnlyPattern()) {
                // InOnly
                // convert Fault to Error if any ...
                logger.finest("original exchange is InOnly");
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                	logger.finest("send back the fault as an Error.");
                	Source faultMsg = newExchange.getFault().getContent();
                	String message = SourceHelper.convertToString(faultMsg);
                	if (message == null) {
                	    message = "an unknown problem was occured. Unable to treat the exchange";
                	}
                	Exception exception = new Exception(message);
                	originalExchange.setError(exception);
                } else {
                    logger.finest("send back nothing.");
                }
                
            } else if (originalExchange.isRobustInOnlyPattern()) {
                // RobustInOnly
                // send back Fault if any
                logger.finest("original exchange is RobustInOnly");
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                	logger.finest("send back the fault.");
                    setFaultOnOriginalExchange(originalExchange, newExchange);
                } else {
                    logger.finest("send back nothing.");
                }
            } else if (originalExchange.isInOptionalOutPattern()) {
                // InOptionalOut
                // send back Fault or Out if any
                logger.finest("original exchange is InOptionalOut");
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                	logger.finest("send back the fault.");
                    setFaultOnOriginalExchange(originalExchange, newExchange);
                } else if (newExchange.isActiveStatus() && newExchange.getOutMessage() != null) {
                	logger.finest("send back the response.");
                	setOutOnOriginalExchange(originalExchange, newExchange);
                } else {
                    logger.finest("send back nothing.");
                }
            } else if (originalExchange.isInOutPattern()) {
                // InOut
                // send back Fault or Out if any, otherwise return an empty Out
                logger.finest("original exchange is InOut");
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                	logger.finest("send back the fault.");
                	setFaultOnOriginalExchange(originalExchange, newExchange);
                } else if (newExchange.isActiveStatus() && newExchange.getOutMessage() != null) {
                	logger.finest("send back the response.");
                	setOutOnOriginalExchange(originalExchange, newExchange);
                } else if (newExchange.isDoneStatus()) {
                	logger.finest("send back a standard response.");
                	setEmptyOutOnOriginalExchange(originalExchange);
                }
            }
        }
    }

    /**
     * Set the original exchange to the Error of the forwarded exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     */
    protected void setErrorOnOriginalExchange(Exchange originalExchange, Exchange newExchange,
            ExchangeContext context) {
        context.getLogger().finest(
                "set on the original exchange an error received from the forwarded exchange.");
        originalExchange.setError(newExchange.getError());
    }

    /**
     * copy on the original exchange the Fault of the forwarded exchange If the
     * fault can not be copied or set, set an error on the original exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    protected void setOutOnOriginalExchange(Exchange originalExchange, Exchange newExchange)
            throws MessagingException, PEtALSCDKException {
        logger
                .finest("set on the original exchange an Out message received from the forwarded exchange.");
        NormalizedMessage out = originalExchange.getOutMessage();
        PatternHelper.copy(newExchange.getOutMessage(), out);
    }

    /**
     * set a default message on the original exchange. set an error on the
     * original exchange if message can not be set
     * 
     * @param originalExchange
     * @param context
     * @throws MessagingException
     */
    protected void setEmptyOutOnOriginalExchange(Exchange originalExchange)
            throws MessagingException {
        Document document = XMLUtil.createDocument();
        Element rootElement = document.createElementNS(EIPConstants.NAMESPACE,
                EIPConstants.RESULT_ELEMENT);
        document.appendChild(rootElement);
        originalExchange.setOutMessageContent(document);

    }

    /**
     * copy on the original exchange the Fault of the forwarded exchange If the
     * fault can not be copied or set, set an error on the original exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * 
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    protected void setFaultOnOriginalExchange(Exchange originalExchange, Exchange newExchange)
            throws MessagingException, PEtALSCDKException {
        logger.finest("set on the original exchange a Fault received from the forwarded exchange.");

        Fault fault = originalExchange.createFault();
        PatternHelper.copy(newExchange.getFault(), fault);
        originalExchange.setFault(fault);

    }

    /**
     * Close the forwarded exchange by sending a DONE status. If the status can
     * not be sent, set an error on the original exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * @throws MessagingException
     */
    protected void closeForwardedExchange(Exchange originalExchange, Exchange newExchange,
            ExchangeContext context) throws MessagingException {
        logger.finest("close the forwarded exchange.");

        newExchange.setDoneStatus();
        context.send(newExchange);

    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "Bridge";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return true;
    }

}
