/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.InputStreamForker;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;

/**
 * Aggregate all Consumes services defined in the service unit. <br>
 * Return the aggregated results.<br>
 * <br>
 * - Accept InOut<br>
 * - If a Fault or Error occurs on a call, stop the process and return the Fault
 * or Error Note : The exchange will be sent back automatically by the framework
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * 
 */
public class ScatterGather extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    public void process(Exchange exchange, ExchangeContext context) throws PEtALSCDKException {
        Exchange originalExchange = exchange;

        List<Consumes> consumesList = context.getSUConsumes(originalExchange.getEndpoint());

        if (consumesList.size() == 0) {
            throw new PEtALSCDKException(
                    "The Service Unit must define at least 1 Consumes sections for the pattern "
                            + this.getPatternName());
        }

        try {
            List<Exchange> sentExchanges = sendAsyncExchanges(consumesList, exchange, context);

            // when messages have been sent, wait for the responses from all
            // services
            receiveResponses(sentExchanges, exchange, context);

        } catch (MessagingException e) {
            throw new PEtALSCDKException(e);
        } catch (InterruptedException e) {
            throw new PEtALSCDKException(e);
        }

    }

    /**
     * Send a copy of the exchange's IN message to all the services to consume.
     * Return the list of the Exchanges sent asynchronously. If an error occurs,
     * set a Fault on the original exchange, and return null TODO if an exchange
     * fails, close the previous sent exchange...
     * 
     * @param consumesList
     * @param exchange
     * @param context
     * @return the list of the exchanges sent async, or null if a problem
     *         occurred
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    private List<Exchange> sendAsyncExchanges(List<Consumes> consumesList, Exchange exchange,
            ExchangeContext context) throws MessagingException, PEtALSCDKException {

        List<Exchange> sentExchanges = new ArrayList<Exchange>();

        NormalizedMessage inMessage = exchange.getInMessage();

        // Handle the Streaming issue
        InputStreamForker inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
        // send all exchanges asynchronously
        for (Consumes consumedService : consumesList) {
            if (inMessageAsStream != null) {
                // Handle the Streaming issue
                Source inSource = new StreamSource(inMessageAsStream.getInputStreamOne());
                inMessageAsStream = new InputStreamForker(inMessageAsStream.getInputStreamTwo());
                inMessage.setContent(inSource);
            }

            MEPConstants mep = PatternHelper.retrievePattern(consumedService.getMep());
            if (mep == null) {
                mep = MEPConstants.valueOf(exchange.getPattern());
            }
            Exchange tmpExchange = PatternHelper.createExchange(consumedService, mep, inMessage,
                    context);

            logSend(consumedService, tmpExchange);
            context.sendAsync(tmpExchange);
            sentExchanges.add(tmpExchange);
        }

        return sentExchanges;
    }

    /**
     * Aggregates all the responses messages and aggregate the resultss.
     * 
     * @param sentExchanges
     * @param originalExchange
     * @param context
     * 
     * @throws MessagingException
     * @throws PEtALSCDKException
     * @throws InterruptedException
     */
    private void receiveResponses(List<Exchange> sentExchanges, Exchange originalExchange,
            ExchangeContext context) throws MessagingException, PEtALSCDKException,
            InterruptedException {

        // build an aggregate document for the result
        Document aggregateDocument = PatternHelper.createAggregateDocument();

        for (Exchange sentExchange : sentExchanges) {
            Exchange aggregateExchange = context.accept(sentExchange);

            NormalizedMessage receivedMessage = PatternHelper.processSentExchange(
                    aggregateExchange, originalExchange, context);
            if (originalExchange.getFault() == null) {
                // aggregate the XML content, attachments and properties
                PatternHelper.aggregateContent(receivedMessage, aggregateDocument);
                PatternHelper.copyAttachments(receivedMessage, originalExchange.getOutMessage());
                PatternHelper.copyProperties(receivedMessage, originalExchange.getOutMessage());
            }
        }
        
        if (originalExchange.getFault() == null) {
            originalExchange.setOutMessageContent(aggregateDocument);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "ScatterGather";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return MEPConstants.IN_OUT_PATTERN.equals(mep);
    }
}
