/**
 * The Splitter EIP pattern
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * @author Adrien LOUIS - EBMWebSourcing
 * 
 */
public class SplitterGather extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "SplitterGather";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public void process(Exchange originalExchange, ExchangeContext context)
            throws MessagingException, PEtALSCDKException {
        final List<Consumes> consumesList = context.getSUConsumes(originalExchange.getEndpoint());

        if (consumesList.size() == 0) {
            throw new PEtALSCDKException("The Service Unit must define only one consumes section");
        }
        final Consumes consumes = consumesList.get(0);

        final String splitterPath = getSplitterPath(context.getExtensions());
        if (StringHelper.isNullOrEmpty(splitterPath)) {
            throw new PEtALSCDKException("The Service Unit must define a splitter-element section");
        }
        logger.finest("Splitter element:" + splitterPath);

        // retrieve the document only once to avoid the Streaming issue
        final Document inMessageContent = originalExchange.getInMessageContentAsDocument();
        originalExchange.getInMessage().setContent(new DOMSource(inMessageContent));
        try {
            NodeList nodeList = PatternHelper.evaluateXPathExpression(splitterPath,
                    inMessageContent);
            logger.finest("Node list generated with " + nodeList.getLength() + " elements");
            final List<Exchange> splittedExchangeList = new ArrayList<Exchange>(nodeList
                    .getLength());

            for (int i = 0; i < nodeList.getLength(); i++) {
                Exchange splittedExchange = PatternHelper.createConsumeExchange(consumes, context);
                if (splittedExchange.getOperation() == null) {
                    splittedExchange.setOperation(originalExchange.getOperation());
                }
                PatternHelper.copy(new DOMSource(nodeList.item(i)),
                        originalExchange.getInMessage(), splittedExchange.getInMessage(), true,
                        true);
                logSend(consumes, splittedExchange);
                context.sendAsync(splittedExchange);
                splittedExchangeList.add(splittedExchange);
                logger.finest("Splitted exchange " + splittedExchange.getExchangeId() + " sent");
            }

            receiveResponses(splittedExchangeList, originalExchange, context);

        } catch (XPathExpressionException e) {
            throw new PEtALSCDKException(e);
        } catch (InterruptedException e) {
            throw new PEtALSCDKException(e);
        }
    }

    /**
     * Receive all the responses and aggregate them (xml content, properties and
     * attachments).
     * 
     * @param sentExchanges
     * @param originalExchange
     * @param context
     * @return
     * @throws MessagingException
     * @throws PEtALSCDKException
     * @throws InterruptedException
     */
    private void receiveResponses(List<Exchange> sentExchanges, Exchange originalExchange,
            ExchangeContext context) throws MessagingException, PEtALSCDKException,
            InterruptedException {

        // build an aggregate document for the result
        Document aggregateDocument = PatternHelper.createAggregateDocument();

        boolean fault = false;
        for (Exchange sentExchange : sentExchanges) {
            Exchange aggregateExchange = context.accept(sentExchange);

            NormalizedMessage receivedMessage = PatternHelper.processSentExchange(
                    aggregateExchange, originalExchange, context);
            if (receivedMessage instanceof Fault) {
                fault = true;
            } else if (!fault) {
                // aggregate the XML content, attachments and properties
                PatternHelper.aggregateContent(receivedMessage, aggregateDocument);
                PatternHelper.copyAttachments(receivedMessage, originalExchange.getOutMessage());
                PatternHelper.copyProperties(receivedMessage, originalExchange.getOutMessage());
            }
        }
        if (originalExchange.getFault() == null) {
            originalExchange.setOutMessageContent(aggregateDocument);
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return MEPConstants.IN_OUT_PATTERN.equals(mep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /**
     * get the splitter element prefix into the extensions
     * 
     * @param extensions
     *            the ConfigurationExtensions used
     * @return The splitter element, or null if not found
     */
    private String getSplitterPath(ConfigurationExtensions extensions) {
        for (String param : extensions.keySet()) {
            if (EIPConstants.PATH.equals(param)) {
                return extensions.get(param);
            }
        }
        return null;
    }

}
