/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.InputStreamForker;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.se.eip.ExchangeContext;

/**
 * Before sending the Exchange to the Service provider (or consumer for the
 * response), Copy the In or OUT message and send it to a "monitoring service" . <br>
 * The message sent to the monitoring service is InOnly <br>
 * request mode : <br>
 * - copy IN to the monitor and send the exchange to the provider; response is
 * received and send back to the consumer <br>
 * response mode : <br>
 * - send the exchange to the provider, receive the response, copy OUT/Fault to
 * the monitor and send back the response to the consumer <br>
 * request-response mode : <br>
 * copy in and out to the monitor
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * 
 */
public class WireTap extends AbstractPattern {

    public static final String WAY = "wiretap-way";

    public static final String REQUEST_WAY = "request";

    public static final String RESPONSE_WAY = "response";

    public static final String REQUESTRESPONSE_WAY = "request-response";

    public static final String REQUEST_ON_RESPONSE_WAY = "request-on-response";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.
     * component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
     */
    public void process(Exchange exchange, ExchangeContext context) throws PEtALSCDKException {
        Exchange originalExchange = exchange;
        List<Consumes> consumesList = context.getSUConsumes(originalExchange.getEndpoint());

        if (consumesList.size() != 2) {
            throw new PEtALSCDKException(
                    "The Service Unit must define 2 Consumes sections for the pattern "
                            + this.getPatternName());
        }

        if (StringHelper.isNullOrEmpty(context.getExtensions().get(WAY))) {
            throw new PEtALSCDKException("Parameter '" + WAY + "' is mandatory for the pattern "
                    + this.getPatternName());
        }
        sendExchange(consumesList, exchange, context);
    }

    /**
     * Send the exchange to its destination and send a copy to the 'monitoring
     * service'. <br>
     * Depending on the "way" parameter , copy the IN or OUT( or fault) message
     * and send it to the 'monitoring service', as InOnly exchange. <br>
     * values for "way" parameter : <br>
     * "request" send the IN message (InOnly MEP) (default mode) <br>
     * "response" send the OUT or Fault or ERROR as an IN message (InOnly MEP)
     * default mode for "way" is "request"
     * 
     * @param consumesList
     * @param exchange
     * @param context
     */
    protected void sendExchange(List<Consumes> consumesList, Exchange exchange,
            ExchangeContext context) {

        try {
            Consumes provider = consumesList.get(0);
            Consumes monitor = consumesList.get(1);

            // send IN to the monitor if Request way is defined
            Exchange monitorInExchange = prepareRequestCopyIfNeeded(monitor, exchange, context);
            String way = context.getExtensions().get(WAY);
            if (monitorInExchange != null) {
                if (REQUEST_WAY.equals(way) || REQUESTRESPONSE_WAY.equals(way)) {
                    logger.info(getPatternName()+ " send a copy of the IN message to ...\n - itf="
                            + monitor.getInterfaceName() + "\n - srv=" + monitor.getServiceName()
                            + "\n - ep =" + monitor.getEndpointName());
                    context.send(monitorInExchange);
                } else {
                    logger
                            .info(getPatternName()+ " keep a copy of the IN message, the copy will be sent after the response is received.");
                }
            }

            // send (sync) exchange to provider
            sendSyncExchange(provider, exchange, context);

            // send IN to the monitor if RequestOnResponse way is defined
            // and response is OK (for InOut or optionalOut response
            if (monitorInExchange != null && REQUEST_ON_RESPONSE_WAY.equals(way)
                    && exchange.isActiveStatus() && exchange.getFault() == null) {
                logger
                        .info(getPatternName()+ " send a copy of the IN message (after OUT response is received) to ...\n - itf="
                                + monitor.getInterfaceName()
                                + "\n - srv="
                                + monitor.getServiceName()
                                + "\n - ep ="
                                + monitor.getEndpointName());
                context.send(monitorInExchange);
            }

            // send OUT, Fault or error to the monitor if Response way is
            // defined
            Exchange monitorOutExchange = prepareResponseCopyIfNeeded(monitor, exchange, context);
            if (monitorOutExchange != null) {
                logger.info(getPatternName()+ " send a copy of the OUT/Fault message to ...\n - itf="
                        + monitor.getInterfaceName() + "\n - srv=" + monitor.getServiceName()
                        + "\n - ep =" + monitor.getEndpointName());
                context.send(monitorOutExchange);
            }

        } catch (JBIException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            exchange.setError(e);
        }
    }

    /**
     * 
     * @param monitor
     * @param exchange
     * @param context
     * @return the exchange for the monitor, null if not needed
     * @throws PEtALSCDKException
     * @throws JBIException
     */
    protected Exchange prepareRequestCopyIfNeeded(Consumes monitor, Exchange exchange,
            ExchangeContext context) throws PEtALSCDKException, JBIException {
        String way = context.getExtensions().get(WAY);
        Exchange monitorExchange = null;
        if (REQUEST_WAY.equals(way) || REQUESTRESPONSE_WAY.equals(way)
                || REQUEST_ON_RESPONSE_WAY.equals(way)) {
            NormalizedMessage inMessage = exchange.getInMessage();

            // Handle the Streaming issue if neccessary
            InputStreamForker inMessageAsStream = PatternHelper.handleStreamSource(inMessage);

            monitorExchange = PatternHelper.createExchange(monitor, MEPConstants.IN_ONLY_PATTERN,
                    inMessage, context);
            // Reset the Source to the InMessage if necessary
            if (inMessageAsStream != null) {
                inMessage.setContent(new StreamSource(inMessageAsStream.getInputStreamTwo()));
            }
        }
        return monitorExchange;
    }

    /**
     * Send the exchange to the provider (only set the information, the
     * framework will send it lately)
     * 
     * @param provider
     * @param exchange
     * @param context
     * @throws PEtALSCDKException
     * @throws JBIException
     */
    protected void sendSyncExchange(Consumes provider, Exchange originalExchange,
            ExchangeContext context) throws PEtALSCDKException, JBIException {
        NormalizedMessage inMessage = originalExchange.getInMessage();

        Exchange providerExchange = PatternHelper.createExchange(provider, MEPConstants
                .valueOf(originalExchange.getExchangePattern()), inMessage, context);

        logSend(provider, providerExchange);
        context.sendSync(providerExchange);

        // process the response
        NormalizedMessage receivedMessage = PatternHelper.processSentExchange(providerExchange,
                originalExchange, context);
        
        if (receivedMessage != null) {
            PatternHelper.copy(receivedMessage, originalExchange.getOutMessage());
        } else if (originalExchange.getFault() == null && originalExchange.isActiveStatus()) {
            // set the done status here to bypass the response way
            originalExchange.setDoneStatus();
        }
    }

    /**
     * 
     * @param monitor
     * @param exchange
     * @param context
     * @return he exchange for the monitor, null if not needed
     * @throws PEtALSCDKException
     * @throws JBIException
     */
    protected Exchange prepareResponseCopyIfNeeded(Consumes monitor, Exchange exchange,
            ExchangeContext context) throws PEtALSCDKException, JBIException {
        String way = context.getExtensions().get(WAY);
        Exchange monitorExchange = null;

        if ((RESPONSE_WAY.equals(way) || REQUESTRESPONSE_WAY.equals(way))
                && exchange.isActiveStatus()) {
            InputStreamForker outMessageAsStream = null;
            InputStreamForker outFaultAsStream = null;

            monitorExchange = context.createConsumeExchange(monitor, MEPConstants.IN_ONLY_PATTERN);

            if (exchange.getFault() != null) {
                // Handle the Streaming issue if necessary
                outFaultAsStream = PatternHelper.handleStreamSource(exchange.getFault());
                PatternHelper.copy(exchange.getFault(), monitorExchange.getInMessage());
            } else {
                // Handle the Streaming issue if necessary
                outMessageAsStream = PatternHelper.handleStreamSource(exchange.getOutMessage());
                PatternHelper.copy(exchange.getOutMessage(), monitorExchange.getInMessage());
            }

            // Handle the Streaming issue if necessary
            if (outMessageAsStream != null) {
                exchange.getOutMessage().setContent(
                        new StreamSource(outMessageAsStream.getInputStreamTwo()));
            } else if (outFaultAsStream != null) {
                exchange.getFault().setContent(
                        new StreamSource(outFaultAsStream.getInputStreamTwo()));
            }
        }
        return monitorExchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "WireTap";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        if (!MEPConstants.IN_OPTIONAL_OUT_PATTERN.equals(mep)) {
            return true;
        }
        return false;
    }
}
